package edu.berkeley.sbp;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.lang.ref.*;
import edu.berkeley.sbp.util.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.*;

/** an element which matches exactly one input token */
public abstract class Atom<T> extends Element implements Topology<T> {

    protected abstract Topology<T> top();
    public    abstract String toString();

    public Topology toAtom() { return this; }

    // Topology Thunks //////////////////////////////////////////////////////////////////////////////

    public Topology<T>       unwrap()                   { return top().unwrap(); }
    public Topology<T>       empty()                    { return top().empty(); }
    public boolean           contains(T v)              { return top().contains(v); }
    public Topology<T>       intersect(Topology<T> t)   { return top().intersect(t); }
    public Topology<T>       minus(Topology<T> t)       { return top().minus(t); }
    public Topology<T>       union(Topology<T> t)       { return top().union(t); }
    public Topology<T>       complement()               { return top().complement(); }
    public boolean           disjoint(Topology<T> t)    { return top().disjoint(t); }
    public boolean           containsAll(Topology<T> t) { return top().containsAll(t); }
    public int               hashCode()                 { return top().hashCode(); }
    public boolean           equals(Object o)           { return o != null && o instanceof Atom && ((Atom)o).top().equals(top()); }

    // Subclasses //////////////////////////////////////////////////////////////////////////////

    public static class Infer<T extends Token> extends Atom<T> {
        private final Element e;
        public Infer(Element e) { this.e = e; }
        public Topology<T> top() { return (Topology<T>)e.toAtom(); }
        public String toString() { return e.toString(); /* FIXME should be toAtom() */ }
    }
    
    public static class Invert<T extends Token> extends Atom<T> {
        private final Atom<T> a;
        public Invert(Atom<T> a) { this.a = a; }
        public Topology<T> top() { return ((Topology<T>)a.top()).complement(); }
        public String toString() { return "~"+a; }
    }
    
    public static class Hack<T extends Token> extends Atom<T> {
        private final Atom<T> a;
        static final Topology leftright =
            edu.berkeley.sbp.misc.CharToken.rightBrace.union(edu.berkeley.sbp.misc.CharToken.leftBrace);
        public Hack(Atom<T> a) { this.a = a; }
        public Topology<T> top() { return ((Topology<T>)a.top()).minus(leftright); }
        public String toString() { return "~"+a; }
    }
    
}

