package edu.berkeley.sbp.misc;
import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import java.lang.ref.*;
import edu.berkeley.sbp.*;
import edu.berkeley.sbp.Token.Location;
import edu.berkeley.sbp.util.*;

public class CharRange extends Atom<CharToken> {
    private String esc(char c) { return StringUtil.escapify(c+"", "[]-~\\\"\'"); }
    private Topology<CharToken> t;
    public CharRange(Topology<CharToken> t) { this.t = t; }
    public Topology<CharToken> top() { return t; }
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append('[');
        Range.Set ranges = ((IntegerTopology)top()).getRanges();
        if (ranges.size() == -1 || ranges.size() > Character.MAX_VALUE/2) {
            sb.append('~');
            ranges = ranges.complement();
        }
        ranges = ranges.intersect(CharToken.all);
        for(Range r : ranges) {
            if (r.isMinNegInf() || r.isMaxPosInf()) throw new Error("should not happen");
            if (r.getMin()==r.getMax()) {
                sb.append(esc((char)r.getMin()));
            } else{
                sb.append(esc((char)r.getMin()));
                sb.append('-');
                sb.append(esc((char)r.getMax()));
            }
        }
        sb.append(']');
        return sb.toString();
    }
    public static final Atom leftBrace  = CharToken.leftBrace;
    public static final Atom rightBrace = CharToken.rightBrace;
    public static Atom set(Range.Set r) { return new CharRange(new IntegerTopology<CharToken>(r)); }

    /** returns an element which exactly matches the string given */
    public static Element string(String s) {
        if (s.length() == 0) return Union.epsilon;
        final String escapified = "\""+StringUtil.escapify(s, "\"\r\n\\")+"\"";
        Element ret;
        if (s.length() == 1) {
            ret =
                new CharRange(new IntegerTopology<CharToken>((int)s.charAt(0))) {
                    public String toString() { return escapified; } };
        } else {
            Union ret2 = new Union("\""+s+"\"_str", true) {
                    public String toString() { return escapified; } };
            Element[] refs = new Element[s.length()];
            for(int i=0; i<refs.length; i++) refs[i] = new CharRange(new IntegerTopology<CharToken>((int)s.charAt(i)));
            ret2.add(Sequence.constant(refs, s, null, null));
            ret = ret2;
        }
        return ret;
    }

}
